/**
  fastcompare.c 

  AUTHOR : Olivier Elemento, Princeton University
  
  Permission is granted to use, modify, distribute this code for any kind of purpose


**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <search.h>
#include <math.h>
#include <malloc.h>


#define max(x,y) (x>y?x:y)
#define min(x,y) (x<y?x:y)

#define MAXSEQ 20000


char*  get_parameter(int argc, char** argv, char* param);
void   chomp(char* s); 
char*  complement(char* s) ;

void   bubbleSort(float *a, int n, int* b) ;
int    nucl(char c); 

double lcumhyper(int i, int s1, int s2, int N); 
double cumhyper(int i, int s1, int s2, int N); 
double bico(int n, int k);
double factln(int n);
double gammln(double xx);
double factrl(int n) ;
void   nrerror(char str[]);
double hypergeom(int i, int s1, int s2, int N);
void   makeOverlap(char* site0, char* site1, char* site2, int size); 
void   saveSpecies1(char* file, char* site1, int size, char** names); 
int CmpFunc(const void* _a, const void* _b);
char *substr(char *string, int start, int length); 
int indexOfKmer(char* kmer); 
int powint(int x, int y); 

typedef struct _prefixNode {
  int next[4];    // A, C, G, T
 } prefixNode;


int addWord(char* s, int m, prefixNode* pnode, int* nextfree); 
int existWord(char* s, int m, prefixNode* pnode); 


int nucl(char c); 


typedef struct _Kmer {
  int   index;      // where is it found
  float score;
} Kmer;


char* nextSequence(FILE* fp, char** name, int* size, int* nextSequence_started, int* nextSequence_ended, char* nextSequence_currentLine); 
char* nextSequence_currentLine; 
int   nextSequence_started = 0;
int   nextSequence_ended = 0;


int verbose = 0;
int scanace = 0;

//void get_overlap(Site* site1, Site* site2, char** names, int* i, int* s1, int* s2);
void get_overlap_array(char* site1, char* site2, int size, char** names, int* i, int* s1, int* s2); 
void saveOverlap(char* file, char* site1, char* site2, int size, char** names); 

int* N;

int main(int argc, char** argv) {
  
  char* fastafile1;
  char* fastafile2;
  char* kmersfile;

  char* mce, *mcb, *c_mce;
  
  int a;

  ENTRY e, *ep, ec, *epc;

  char** sites_species1;
  char** sites_species2;
  char** sites_species1_overlaps = 0;

  char** names_species1;


  char** kmers;
  
  int kmer_size;
  
  int    i, s, nn, aa, bb;
  int    overlap_i, overlap_s1, overlap_s2;

  FILE*  fp1, *fp2, *fp;
  char*  name1;
  char*  name2;
  int    size1;
  int    size2;
  char*  seq1;
  char*  seq2;
  
  int    n=0, l, ii, j, jj;

  char*  buff;

  int    nbgenes;


  float* scores;
  int*   a_s1;
  int*   a_s2;
  int*   a_i;
  int*   indexes;   // holds initial indices
  int    i_limit = 0;

  float  myscore; 
  char*  file;
  
  
  int    gap = 0;
  int    motifsize = 7;
  char*  overlap_dir;
  char*  compare_file;
  char   stmp[50];

  int    compare = 0;
  int    overlap = 0;

  int   nextSequence_started1 = 0;
  int   nextSequence_ended1   = 0;
  char* nextSequence_currentLine1; 
 
  
  int   nextSequence_started2 = 0;
  int   nextSequence_ended2   = 0;
  char* nextSequence_currentLine2; 

  unsigned char* local_sites1 = 0;
  unsigned char* local_sites2 = 0;

  Kmer* kmers_array;


  // PREFIX VARS
  prefixNode* pnode;
  int nextfree;
  int node;
 
  int last_nextfree;
  int lastnode;
  int* nodes;
  int* node2index;
  int nb1, nb2;

  int singlestrand = 0;

  int nbcopies     = 1;

  int* posinvector;   // new index kmer => i in kmers[i]

  unsigned short* local_last_position_sites;


  
  N = (int*)calloc(256, sizeof(int));
  
  N['A'] = 0;
  N['C'] = 1;
  N['T'] = 2;
  N['G'] = 3;

  if (argc < 2) {
    printf("Usage : fastcompare -kmers FILE -fasta1 FILE -fasta2 FILE -nbgenes INT -k INT [ -gap INT -limit INT -singlestrand [TF] ]\n");
    exit(0);
  }

  nextSequence_currentLine1 = (char*)malloc(50000 * sizeof(char));
  nextSequence_currentLine2 = (char*)malloc(50000 * sizeof(char));


  if (strcmp(get_parameter(argc, argv, "-verbose"), "T") == 0) 
    verbose = 1;

  if (strcmp(get_parameter(argc, argv, "-singlestrand"), "T") == 0) 
    singlestrand = 1;


  kmersfile  = get_parameter(argc, argv, "-kmers");
  fastafile1 = get_parameter(argc, argv, "-fasta1");
  fastafile2 = get_parameter(argc, argv, "-fasta2");

  
  if (strlen(get_parameter(argc, argv, "-nbgenes")) == 0) 
    nbgenes = 6200;
  else
    nbgenes = atoi(get_parameter(argc, argv, "-nbgenes"));

  
  if (strlen(get_parameter(argc, argv, "-limit")) == 0) 
    i_limit = 0;
  else
    i_limit = atoi(get_parameter(argc, argv, "-limit"));


  if (strlen(get_parameter(argc, argv, "-gap")) == 0) 
    gap = 0;
  else
    gap = atoi(get_parameter(argc, argv, "-gap"));

   if (strlen(get_parameter(argc, argv, "-nbcopies")) == 0) 
    nbcopies = 1;
  else
    nbcopies = atoi(get_parameter(argc, argv, "-nbcopies"));

   
  
   


  if (strlen(get_parameter(argc, argv, "-verbose")) == 0) 
    verbose = 0;
  else
    verbose = 1;

  if (strlen(get_parameter(argc, argv, "-k")) == 0) 
    kmer_size = -1;
  else
    kmer_size = atoi(get_parameter(argc, argv, "-k"));



  overlap_dir  = get_parameter(argc, argv, "-overlap_dir");
  compare_file = get_parameter(argc, argv, "-compare_file");

  compare = 0;
  if (strlen(compare_file) > 0) {
    compare = 1;
  }

  overlap = 0;
  if (strlen(overlap_dir) > 0) {
    overlap = 1;
  }
  

   if (verbose) {
     printf("Nb copies = %d\n", nbcopies);
   }

  //
  // READ THE LIST OF KMERS
  //


  if (kmer_size == 0) {
    kmer_size = 10;
  } 

  //n = (int)pow(4, kmer_size);
  //else 
  //n = 100000;

  if (kmer_size > 10) {
    printf("Can't deal with such large k-mers ..\n");
    exit(0);
  }
  
  n           = indexOfKmer(substr("GGGGGGGGGGGGG", 0, kmer_size)) + 1;  
  posinvector = (int*)malloc(n * sizeof(int));
  for (i=0; i<n; i++) {
    posinvector[i] = -1;
  }


  kmers = (char**)malloc(n * sizeof(char*));

  // create a line buffer
  buff = (char*)malloc(10000 * sizeof(char));
  fp = fopen(kmersfile, "r");
  if (!fp) {
    printf("could not open set data %s\n", kmersfile);
  }

  n = 0;
  motifsize = 0;
  while (!feof(fp)) {
    fscanf(fp, "%s\n", buff);
    kmers[n] = (char*)calloc(100, sizeof(char));
    
    strcat(kmers[n], buff);
    
    if (motifsize < strlen(kmers[n])) {
      motifsize = strlen(kmers[n]);
    }

    // create a new entry
    posinvector[ indexOfKmer(kmers[n]) ] = n;

    if (singlestrand == 0) {
      c_mce = complement(kmers[n]);
      posinvector[ indexOfKmer(c_mce)    ] = n;
      free(c_mce);
    }

    n++;

  }
  
  fclose(fp);
  
  sites_species1 = (char**)malloc(n * sizeof(char*));
  sites_species2 = (char**)malloc(n * sizeof(char*));
  
  // alloc memory for local storage
  local_sites1    = (char*)malloc(n * sizeof(char));
  local_sites2    = (char*)malloc(n * sizeof(char));
  
  // store last occurences
  local_last_position_sites    = (unsigned short*)malloc(n * sizeof(short));
    

  // score storages
  scores         = (float*)malloc(n * sizeof(float));
  a_s1           = (int*)calloc(n, sizeof(int));
  a_s2           = (int*)calloc(n, sizeof(int));
  a_i            = (int*)calloc(n, sizeof(int));
  indexes        = (int*)malloc(n * sizeof(int));


  

    
  /*
   *
   *  BUILD A PREFIX TREE  KMER => INDEX in sites_speciesX
   *
   */
  

 
     
    
  // associates a node to an index
  //node2index = (int*)malloc(n * motifsize * sizeof(int));
  
  // we are going to need many nodes ...
  //pnode = (prefixNode*)malloc(motifsize * n * (sizeof(prefixNode)));
  //pnode[0].next[0] = -1;
  //pnode[0].next[1] = -1;
  //pnode[0].next[2] = -1;
  //pnode[0].next[3] = -1;


  //
  //  add all the words in the list
  //
  //nextfree = 1;

  //for (i=0; i<n; i++) {
  //lastnode = addWord(kmers[i], motifsize, pnode, &nextfree);
  
  //
  //   at this point, lastnode contains the last node inserted in the tree
  //
  
  //node2index[lastnode] = i;
  //}
  
  
    
  //
  //  open sequence file 1
  //
  fp1 = fopen(fastafile1, "r");
  if (!fp1) {
    printf("cannot open %s ..\n", fastafile1);
    exit(0);
  }

  //
  //  open sequence file 2
  //
  fp2 = fopen(fastafile2, "r");
  if (!fp2) {
    printf("cannot open %s ..\n", fastafile2);
    exit(0);
  }
    
  
  mce   = (char*)calloc(motifsize+2, sizeof(char));
  mcb   = (char*)calloc(motifsize+2, sizeof(char));
  c_mce = (char*)calloc(motifsize+2, sizeof(char));


  //
  //
  //
  //  SCAN SPECIES 1 AND 2
  //
  //


  s = 0;
  while ( (seq1 = nextSequence(fp1, &name1, &size1, &nextSequence_started1, &nextSequence_ended1, nextSequence_currentLine1)) ) {   

    //printf("Read SEQ 1 %s, %s ..\n", name1, seq1); 

        
    l = strlen(seq1);

    //  initilialize the local store
    memset(local_sites1, 0, n);

    // memorize the position of the last kmer
    memset(local_last_position_sites, 0, n * sizeof(unsigned short));

    
    for (j=0; j<strlen(seq1) - motifsize - gap + 1; j++) {
      
      if (gap > 0) {
	strncpy(mce, seq1+j,  (motifsize / 2));
	strncpy(mce+(motifsize / 2), seq1+j + (motifsize / 2) + gap, motifsize - (motifsize / 2));
      } else {
	strncpy(mce, seq1+j, motifsize);
      }


      mce[motifsize] = '\0';
      //c_mce          = complement(mce);
      

      nb1 = posinvector[ indexOfKmer(mce) ]; //existWord(mce, motifsize, pnode);

      //nb2 = -1;
      //if (singlestrand != 0)
      //nb2 = posinvector[ indexOfKmer(c_mce) ]; //existWord(c_mce, motifsize, pnode);

      
      // if any of them has been found, no need to go on
      //if ((nb1 > 0) || (nb2 > 0)) {
	
      if (nb1 >= 0) {

	i = nb1; //node2index[nb1];
	//else
	//i = node2index[nb2];
	
	// if the current motif does not overlap with the last same one
	if ((j != 0) || (local_last_position_sites[i]+motifsize+gap <= j)) {

	  
	  // the ith kmer is present in this upstream region
	  if (local_sites1[i] <= 255)
	    local_sites1[i] ++; 	  
	  
	  
	  
	  // increment s1 for the ith kmer if it has not already been incremented
	  if (local_sites1[i] == nbcopies) {
	    a_s1[i] ++;
	    
	    
	  }

	  local_last_position_sites[i] = j;

	}
	
	
	
	
      }
      
      // free memory
      //free(c_mce);
    }
    
    
    free(name1);
    free(seq1);

    //
    // SCAN SEQUENCE 2
    //

    seq2 = nextSequence(fp2, &name2, &size2, &nextSequence_started2, &nextSequence_ended2, nextSequence_currentLine2);

    // printf("Read SEQ 2 %s, %s ..\n", name2, seq2); 
    //  initilialize the local store
    memset(local_sites2, 0, n);
 
    // memorize the position of the last kmer
    memset(local_last_position_sites, 0, n * sizeof(unsigned short));


    
    l = strlen(seq2);
    
    for (j=0; j<strlen(seq2) - motifsize - gap + 1; j++) {
      
      if (gap > 0) {
	strncpy(mce, seq2+j,  (motifsize / 2));
	strncpy(mce+(motifsize/2), seq2+j + (motifsize / 2) + gap, motifsize - (motifsize / 2));
      } else {
	strncpy(mce, seq2+j, motifsize);
      }
      
      mce[motifsize] = '\0';
      //c_mce = complement(mce);
      
      nb1 = posinvector[ indexOfKmer(mce) ]; //existWord(mce, motifsize, pnode);
      
      //nb2 = -1;
      //if (singlestrand != 0)
      //nb2 = existWord(c_mce, motifsize, pnode);
      
      
      // if any of them has been found, no need to go on
      //if ((nb1 > 0) || (nb2 > 0)) {
	
      if (nb1 >= 0) {
	i = nb1; //node2index[nb1];
	//else
	//i = node2index[nb2];

	
	if ((j != 0) || (local_last_position_sites[i]+motifsize+gap <= j)) {
	  
	  
	  if (local_sites2[i] <= 255)
	    local_sites2[i] ++; 
	  
	  // I just changed the i box
	  if (local_sites2[i] == nbcopies) {
	    
	    // increment s2 for this kmer 
	    a_s2[i] ++;
	    
	    // the ith kmer is present in this upstream region
	    //local_sites2[i] = 1; 	  
	    //if (local_sites2[i] <= 255)
	    //local_sites2[i] ++; 	
	    
	    // the same kmer was present in the orth upstream sequence
	    if (local_sites1[i] >= nbcopies) {
	      a_i[i] ++;
	    } 
	  }
	  
	  local_last_position_sites[i] = j;

	  
	}
	
	

      }
      
      //free(c_mce);
    }
    
    
    free(name2);
    free(seq2);


    s++;

    if (s == 50) {
      //break;
    }

  }

  fclose(fp1);
  fclose(fp2);


  // no need to keep the HASH table
  hdestroy();



  //
  // traverse every kmers
  //

  kmers_array = (Kmer*)malloc(n * sizeof(Kmer));
  
  for (i=0; i<n; i++) {

    //printf("Calc score for kmer %s, i=%d, s1=%d, s2=%d, N=%d\n", kmers[i], a_i[i], a_s1[i], a_s2[i], nbgenes); getchar();
  
    // calculate (and store)  conservation score for this kmer

    kmers_array[i].index = i;
    kmers_array[i].score = lcumhyper(a_i[i], a_s1[i], a_s2[i], nbgenes);

    


  }


  //printf("sort ..");

  // sort the kmers by their score (inefficient bubble sort for now) 
  //bubbleSort(scores, n, indexes);  
  qsort((void*)kmers_array, n, sizeof(Kmer), CmpFunc);
  
  //printf(" done\n");

  a = 0;
  for (i=0; i<n; i++) {
        
    // get index in initial array
    ii = kmers_array[i].index;
  

    // output the kmer itself (with gap if needed)
    if (gap > 0) {
      strncpy(stmp, kmers[ii], motifsize/2);  
      stmp[motifsize/2] = '\0';
      printf("%s", stmp);	    
      for (j=0; j<gap; j++) 
	printf("N");
      printf("%s\t", kmers[ii] + motifsize/2);	    
    } else {
      printf("%s\t", kmers[ii]);
    }	



    // output s1, s2, i and score for this kmer
    printf("%d\t%d\t%d\t%5.4f\n", a_s1[ii], a_s2[ii],  a_i[ii], - kmers_array[i].score);
    
    
    
    a++;
    
    // abort the output if the user asked for the highest scoring kmers
    if ( (i_limit > 0) && (i_limit == a)) {
	break;
    }
  }



  

  for (i=0; i<n; i++) {
    free(kmers[i]);
  }
  free(kmers);
  free(buff);
  free(mce);

  
  free( nextSequence_currentLine1);
  free( nextSequence_currentLine2);
    
  return 0;
}


int powint(int x, int y) 
{
  
  int o = 1;
  int i = 0;

  for (i=0; i<y; i++) {
    o *= x; 
  }
  
  return o;
}


int indexOfKmer(char* kmer) 
{
  
  //int* N;
  int idx = 0;
  int i;
  int l   = strlen(kmer);

  
  

  for (i=0; i<l; i++) {
    idx += N[ (int)(kmer[i]) ] * powint(4,(l - i - 1)); 
  }
  
  //free(N);
  
  return idx;
}


//
//  substring
//
char *substr(char *string, int start, int length) 
{
  char *substring;
  
  substring = (char*)calloc(length+1, sizeof(char));

  memcpy(substring, string+start, length*sizeof(char));
  substring[length] = '\0';

  return substring;
}




//
// 
//   tri simple de type bubble sort
//        *a is the array of scores to sort
//        *b correspond to the initial index of the kmers list !
//
void bubbleSort(float *a, int n, int* b) 
{
  int i, j;

  // for storing temporary values
  int   itmp;
  float ftmp;
  
  
  for (i=0; i<n-1; i++) {
    for (j=0; j<n-1-i; j++)
      if (a[j+1] > a[j]) {  /* compare the two neighbors */
        
	ftmp = a[j];         /* swap a[j] and a[j+1]      */
        a[j] = a[j+1];
        a[j+1] = ftmp;

	itmp = b[j];         /* swap a[j] and a[j+1]      */
        b[j] = b[j+1];
        b[j+1] = itmp;

      }
  }

}
//





//
//  get the overlap between two arrays of 0,1
//    ie i, unique s1, unique s2
//  size = number of orthologs
void get_overlap_array(char* site1, char* site2, int size, char** names, int* i, int* s1, int* s2) 
{

  int overlap_i = 0, overlap_s1 = 0, overlap_s2 = 0;
  int ii;
    
  
  for (ii=0; ii<size; ii++){
    if ( (site1[ii] == 1) && (site2[ii] == 1) ) {
      overlap_i++;
      if (verbose)
	printf("%s\n", names[ii]);
    } 
    
    if (site1[ii] == 1) {
      overlap_s1++;
    }
    
    if (site2[ii] == 1) {
      overlap_s2++;
    }
    
  }
  
  *i = overlap_i;
  *s1  = overlap_s1;
  *s2 = overlap_s2;


   
}

//
//  save the overlap to a file
//
void saveOverlap(char* file, char* site1, char* site2, int size, char** names) 
{
  
  int ii;
  FILE* fp;


  fp = fopen(file, "w");

  
  for (ii=0; ii<size; ii++){
    if ( (site1[ii] == 1) && (site2[ii] == 1) ) {
      fprintf(fp, "%s\n", names[ii]);
    } 
  }

  fclose(fp);
  
}



//
//  save the overlap to a file
//
void saveSpecies1(char* file, char* site1, int size, char** names) 
{
  
  int ii;
  FILE* fp;


  fp = fopen(file, "w");

  
  for (ii=0; ii<size; ii++){
    if (site1[ii] == 1) {
      fprintf(fp, "%s\n", names[ii]);
    } 
  }

  fclose(fp);
  
}








//
//  save the overlap to a file
//
void makeOverlap(char* site0, char* site1, char* site2, int size) 
{
  
  int ii;
  
  for (ii=0; ii<size; ii++){
    if ( (site1[ii] == 1) && (site2[ii] == 1) ) {
      site0[ii] = 1;
    } 
  }

  
}

char* get_parameter(int argc, char** argv, char* param)
{
  int i = 0;
  while ((i < argc) && (strcmp(param, argv[i])))
    i++;
  if (i<argc)
    return (argv[i+1]);
  else
    return "";
}



void chomp(char* s) 
{

  int j;

  j = strlen(s);
  while (j >= 0) {
    if (s[j] == '\n') {
       s[j] = '\0';
       return;
    }
    j--;
  }


}







char* complement(char* s) 
{

  int l = strlen(s);
  char* c;
  char  d;
  int i;
  
  c = (char*)calloc(l+1, sizeof(char));

  for (i=l-1; i>=0; i--) {
    if (s[i] == 'A')
      d = 'T';
    else if (s[i] == 'T')
      d = 'A';
    else if (s[i] == 'G')
      d = 'C';
    else if (s[i] == 'C') 
      d = 'G';
    else 
      d = 'N';
    
    strncat(c, &d, 1);
  }
	 
  return c;
  
}

//
//  returns the next sequence in the file fp
//
char* nextSequence(FILE* fp, char** name, int* size, int* nextSequence_started, int* nextSequence_ended, char* nextSequence_currentLine) 
{
  
  int len = 20000;
  char* seq;
  

  // not yet started, get the first line
  if (*nextSequence_started == 0) {
    fgets(nextSequence_currentLine, len, fp);  
    chomp(nextSequence_currentLine);      
    *nextSequence_started = 1;
 
    
 }
  
  //printf("%s\n", nextSequence_currentLine);

  // if file is finished, return 0
  if (*nextSequence_ended == 1) {
    return 0;
  }

  // if started, line should be filled with ">..."
  if (nextSequence_currentLine[0] == '>') {
    

    *name = strdup(nextSequence_currentLine + 1);
    
    // create a new line
    seq = (char*)calloc(100000, sizeof(char) ); 

    while (1) {
      
      fgets(nextSequence_currentLine, len, fp);    
      
      if (feof(fp)) {
	*nextSequence_ended  = 1;
	return seq;
      }

      chomp(nextSequence_currentLine);      
      if (strlen(nextSequence_currentLine) == 0) { 
	continue;
      }
      
      if (nextSequence_currentLine[0] == '>') {	
	return seq;	
      } else {
	strcat(seq, nextSequence_currentLine);
      }
      
    }

  } else return 0;
  
}




double lcumhyper(int i, int s1, int s2, int N) {
  
  return log(cumhyper(i, s1, s2, N));

}


double cumhyper(int i, int s1, int s2, int N) {
  
  int min = (s1<s2?s1:s2);
  double prod = 0.0;
  int a;
  double tmp = 0.0;

  for (a=i; a<=min; a++) {
    tmp = (double)hypergeom(a, s1, s2, N);
    prod += (double)hypergeom(a, s1, s2, N);
  }
  
  return prod;
}



void nrerror(char str[]) {
  printf("%s\n", str);
}


double hypergeom(int i, int s1, int s2, int N) {
  double factln(int n);
    
  return exp(factln(s1) + factln(N - s1) + factln(s2) + factln(N - s2) - 
	     factln(i) - factln(N) - factln(s1 - i) - factln(s2 - i) 
	     - factln(N - s1 - s2 + i));
    
}



double factln(int n) {
  double gammln(double xx);
  void nrerror(char error_text[]);
  static double a[101];                                        
  if (n < 0) nrerror("Negative factorial in routine factln");
  if (n <= 1) return 0.0;
  if (n <= 100) return a[n] ? a[n] : (a[n]=gammln(n+1.0));                                    
  else return gammln(n+1.0);                                  
}


double gammln(double xx) {
  double x,y,tmp,ser;
  static double cof[6]={76.18009172947146,-86.50532032941677,
			24.01409824083091,-1.231739572450155,
			0.1208650973866179e-2,-0.5395239384953e-5};
  int j;
  y=x=xx;
  tmp=x+5.5;
  tmp -= (x+0.5)*log(tmp);
  ser=1.000000000190015;
  for (j=0;j<=5;j++) ser += cof[j]/++y;
  return -tmp+log(2.5066282746310005*ser/x);
}



double factrl(int n)  { 
  double gammln(double xx); 
  void nrerror(char error_text[]); 
  static int ntop=4; 
  static double a[33]={1.0,1.0,2.0,6.0,24.0}; 
  int j; 
  
  if (n < 0) 
    nrerror("Negative factorial in routine factrl"); 
  if (n > 32) 
    return exp(gammln(n+1.0)); 
  while (ntop<n) { 
    j=ntop++; 
    a[ntop]=a[j]*ntop; 
  } 
  
  return a[n]; 

}



int CmpFunc(const void* _a, const void* _b)
{
  const Kmer* a = (const Kmer*) _a;
  const Kmer* b = (const Kmer*) _b;
  
  if (a->score > b->score) 
    return 1; 
  else if(a->score == b->score) 
    return  0;
  else         
    return -1;
}




/*** PREFIX TREE OPTIONS ***/

int nucl(char c) 
{
  
  if (c == 'A')
    return 0;
  else if (c == 'C')
    return 1;
  else if (c == 'G')
    return 2;
  else if (c == 'T')
    return 3;
  else return -1;
  
  
}






// add word to the prefix tree (if it is not there)
int addWord(char* s, int m, prefixNode* pnode, int* nextfree) 
{

  //printf("Adding %s to the prefix tree ..\n", s);
  int i, k;

  int j = 0;

  for (i=0; i<m; i++) {
    k = pnode[j].next[nucl(s[i])];
    if (k == -1) {
      pnode[j].next[nucl(s[i])] = *nextfree;

      pnode[*nextfree].next[0] = -1;
      pnode[*nextfree].next[1] = -1;
      pnode[*nextfree].next[3] = -1;
      pnode[*nextfree].next[2] = -1;

      (*nextfree)++;
    }
    j = pnode[j].next[nucl(s[i])];
  }

  return j;
}


// add word to the prefix tree (if it is not there)
int existWord(char* s, int m, prefixNode* pnode) 
{

  //printf("Adding %s to the prefix tree ..\n", s);
  int i, k;
  int t;
  int j = 0;

  for (i=0; i<m; i++) {
    t = nucl(s[i]);
    
    if (t == -1)
      return -2;

    k = pnode[j].next[nucl(s[i])];
    if (k == -1)
      return -1;
  
    j = pnode[j].next[nucl(s[i])];

  }

  return j;
}
